#include <u.h>
#include <libc.h>
#include <bio.h>

typedef struct Vendor Vendor;
typedef struct Product Product;
typedef struct Class Class;
typedef struct Subclass Subclass;

struct Vendor
{
	ulong	id;
	char	*abbrev;
	char	*name;
};

struct Product
{
	ulong	id;
	char	*abbrev;
	char	*name;
	ulong vid;
};

struct Class
{
	int	id;
	char	*name;
};

struct Subclass
{
	int	class;
	int	id;
	char *name;
};


Vendor *vendors;
Product *products;
Class *classes;
Subclass * subclasses;

int nvendors;
int nproducts;
int nclasses;
int nsubclasses;


char *dbfile = "pcidb";
char *defsfile = "pcidefs";
char *pcidir = "/dev/pci";


void
usage(void)
{
	fprint(2, "usage: pcidevs\n");
	exits("usage");
}


char *
estrdup(char *s)
{
	char *p;
	p = malloc(strlen(s) +1);
	if(!p)
		sysfatal("malloc");
	strcpy(p, s);
	return p;
}


void *
erealloc(void *p, ulong size)
{
	p = realloc(p, size);
	if(!p)
		sysfatal("realloc");
	return p;
}


ulong
getvendorid(char *abbrev)
{
	int i;
	for (i = 0; i < nvendors; i++)
		if(strcmp(vendors[i].abbrev, abbrev) == 0)
			return vendors[i].id;
	return -1;
}


char *
getvendor(ulong id)
{
	int i;
	for(i = 0; i < nvendors; i++)
		if(vendors[i].id == id)
			return vendors[i].name;
	return "Unknown";
}


char *
getproduct(ulong vid, ulong did)
{
	int i;
	for(i = 0; i < nproducts; i++)
		if(products[i].vid == vid && products[i].id == did)
			return products[i].name;
	return "Unknown";
}

char *
getclass(int class)
{
	int i;
	for(i = 0; i < nclasses; i++) {
		if (classes[i].id == class)
			return classes[i].name;
	}
	return "Unknown";
}

char *
getsubclass(int class, int subclass)
{
	int i;
	for(i = 0; i < nsubclasses; i++)
		if(subclasses[i].class == class && subclasses[i].id == subclass)
			return subclasses[i].name;
	return "Unknown";
}


void
readdb(void)
{
	Biobuf *b;
	char *l;
	char *fields[5];

	b = Bopen(dbfile, OREAD);
	if (!b)
		sysfatal("opening %s: %r", dbfile);

	while(l = Brdstr(b, '\n', 1)) {
		if (strncmp(l, "vendor", strlen("vendor")) == 0) {
			if (getfields(l, fields, 4, 1, " \t") != 4) {
				fprint(2, "ignoring vendor line: %s\n", l);
				free(l);
				continue;
			}

			vendors = erealloc(vendors, sizeof vendors[0] * (nvendors+1));
			vendors[nvendors++] = (Vendor){strtoul(fields[2], nil, 16), estrdup(fields[1]), estrdup(fields[3])};
		}
		if(strncmp(l, "product", strlen("product")) == 0) {
			if(getfields(l, fields, 5, 1, " \t") != 5) {
				fprint(2, "ignoring product line: %s\n", l);
				free(l);
				continue;
			}

			products = erealloc(products, sizeof products[0] * (nproducts+1));
			products[nproducts++] = (Product){strtoul(fields[3], nil, 16), estrdup(fields[2]), estrdup(fields[4]), getvendorid(fields[1])};
		}
		free(l);
	}

	Bterm(b);
}

void
readdefs(void)
{
	Biobuf *b;
	char *l;
	char *fields[4];

	b = Bopen(defsfile, OREAD);
	if(!b)
		sysfatal("opening %s: %r", defsfile);

	while(l = Brdstr(b, '\n', 1)) {
		if(strncmp(l, "class", strlen("class")) == 0) {
			if(getfields(l, fields, 3, 1, " \t") != 3) {
				fprint(2, "ignoring class line: %s\n", l);
				free(l);
				continue;
			}
			classes = erealloc(classes, sizeof classes[0] * (nclasses+1));
			classes[nclasses++] = (Class){atoi(fields[1]), estrdup(fields[2])};
		}
		if (strncmp(l, "subclass", strlen("subclass")) == 0) {
			if(getfields(l, fields, 4, 1, " \t") != 4) {
				fprint(2, "ignoring subclass line: %s\n", l);
				free(l);
				continue;
			}
			subclasses = erealloc(subclasses, sizeof subclasses[0] * (nsubclasses+1));
			subclasses[nsubclasses++] = (Subclass){atoi(fields[1]), atoi(fields[2]), estrdup(fields[3])};
		}
		free(l);
	}

	Bterm(b);
}


void
main(int argc, char *argv[])
{
	Dir *pcifiles;
	int npcifiles;
	int fd;
	char *filename;
	char line[19];
	char *fields[5];
	int i;
	char *type;


	ARGBEGIN {
	default:
		usage();
	} ARGEND

	readdb();
	readdefs();

	fd = open(pcidir, OREAD);
	if (fd < 0)
		sysfatal("opening %s: %r", pcidir);
	npcifiles = dirreadall(fd, &pcifiles);
	if (npcifiles < 0)
		sysfatal("reading %s: %r", pcidir);
	close(fd);

	print("BDF     Class.Sub.If                         Vendor/Product\n");

	for (i = 0; i < npcifiles; i++) {
		Dir *pcifile = pcifiles+i;

		if (strstr(pcifile->name, "ctl")) {
			int class, subclass, prog;
			ulong vid, did;

			filename = smprint("%s/%s", pcidir, pcifile->name);
			fd = open(filename, OREAD);
			if (fd < 0)
				sysfatal("opening %s: %r", filename);

			/* e.g. 06.00.00 1106/3123, 18 chars */
			if (read(fd, line, 18) != 18)
				sysfatal("reading %s", filename);

			pcifile->name[strlen(pcifile->name)-3] = '\0';

			if (getfields(line, fields, 5, 1, " ./") != 5) {
				fprint(2, "pcidevs: ignoring line: %s\n", line);
				free(filename);
				continue;
			}

			class = strtoul(fields[0], nil, 16);
			subclass = strtoul(fields[1], nil, 16);
			prog = strtoul(fields[2], nil, 16);
			vid = strtoul(fields[3], nil, 16);
			did = strtoul(fields[4], nil, 16);

			type = smprint("(%s/%s)", getclass(class), getsubclass(class, subclass));

			print("%6s  %02x.%02x.%02x  %-25s  %04lux/%04lux (%s, %s) \n", pcifile->name, class, subclass, prog, type, vid, did, getvendor(vid), getproduct(vid, did));

			close(fd);
			free(filename);
		}
	}

	free(pcifiles);
	exits(0);
}
